#!/usr/bin/python

import random
import copy
import sys

kindToCoor = {"O":[(0,0),(0,1),(1,0),(1,1)],"T":[(1,0),(1,1),(1,2),(2,1)],
    "F":[(1,0),(1,1),(1,2),(2,0)],"L":[(1,0),(1,1),(1,2),(0,0)],
    "I":[(1,0),(1,1),(1,2),(1,3)],"S":[(0,0),(0,1),(1,1),(1,2)],
    "Z":[(1,0),(1,1),(0,1),(0,2)]}

kindToColour = {"O":"Blue","T":"Purple","F":"Red","L":"Green","I":"Yellow",
    "S":"Orange","Z":"White"}

class Piece:
    "A Piece contains a _kind, _colour, _coor, _xoffset, _yoffset"
    
    def __init__(self):
        # The offsets are used for the rotation.
        self._xoffset = self._yoffset = 1
        self._kind = random.choice(["O","T","L","F","I","S","Z"])
        self._coor = kindToCoor[self._kind]
        self._colour = kindToColour[self._kind]

    def movePiece(self,dir):
        if dir == "Left":
            self._coor = [(x-1,y) for (x,y) in self._coor]
            self._xoffset = self._xoffset - 1
        elif dir == "Right":
            self._coor = [(x+1,y) for (x,y) in self._coor]
            self._xoffset = self._xoffset + 1
        elif dir == "Down":
            self._coor = [(x,y+1) for (x,y) in self._coor]
            self._yoffset = self._yoffset + 1
        elif dir == "Rotate":
            self._coor = [(self._xoffset - self._yoffset + y,
                self._xoffset + self._yoffset - x) for (x,y) in self._coor]

    def checkBounds(self):
        "Return true if the piece is within the bounds of the board."
        
        for (x,y) in self._coor:
            if x < 0 or x > 10 or y < 0 or y > 17:
                return False
        return True

    def checkBottom(self):
        "Return true if the piece is at the bottom of the board."

        for (x,y) in self._coor:
            if y > 17: return True
        return False

class Board:
    "A Board is simply a dictionary from (int,int) -> colour"

    def __init__(self):
        self._board = dict([((x,y),"None")
            for x in range(0,11) for y in range(0,18)])
    
    def putPiece(self,piece):
        """Puts the piece on the board (i.e. changes the coordinates of the
        piece on the board to be the colour of the piece."""

        for coor in piece._coor:
            self._board[coor] = piece._colour

    def checkCollide(self,piece):
        "Returns true if this piece collides with another on the board."

        for coor in piece._coor:
            if self._board[coor] != "None": return True
        return False
    
    def eraseLines(self):
        "Erases full lines on the board."
        
        for row in range(0,18):
            # Assume the row is full
            rowfull = True
            for coor in [(x,row) for x in range(0,11)]:
                # If any of the elements in the row are None, the row is
                # not full and we break the loop to check the next row.
                if self._board[coor] == "None":
                    rowfull = False
                    break
            
            if rowfull:
                # If the row is full, copy the values in row-1 down to row
                # Row 0 will not be copied on to, it will be filled with None's
                for (x,y) in [(x,y)
                    for y in range(row,0,-1) for x in range(0,11)]:
                    self._board[(x,y)] = self._board[(x,y-1)]

                #Fill the top row with "None"s
                for coor in [(x,0) for x in range(0,11)]:
                    self._board[coor] = "None"

class GameState:
    "A GameState has a _board and a currently active _piece"

    def __init__(self):
        self._score = self._lines = 0
        self._level = 1
        self._board = Board()
        self._piece = Piece()

    def makeMove(self,dir):
        "Given a direction, it checks if the move is valid and then moves."
        
        # Make a copy of the current piece and move it (to do the tests)
        movedPiece = copy.deepcopy(self._piece)
        movedPiece.movePiece(dir)
        
        if dir == "Down" and (movedPiece.checkBottom() or self._board.checkCollide(movedPiece)):
            self._board.putPiece(self._piece)
            self._board.eraseLines()
            self._lines = self._lines + 1
            self._score = self._score + 1
            self._level = self._level + 1
            self._piece = Piece()
        elif dir == "Down" or (movedPiece.checkBounds() and
            not self._board.checkCollide(movedPiece)):
            self._piece.movePiece(dir)

    def printit(self):
        boardWithPiece = copy.deepcopy(self._board)
        boardWithPiece.putPiece(self._piece)
        for i in range(0,18):
            for j in range(0,11):
                colour = boardWithPiece._board[(j,i)]
                if colour == "None": letter = "-"
                else: letter = "#"
                sys.stdout.write(letter)
            print ""

